#include <iostream>
#include <fstream>
#include <list>
#include <sys/time.h>
#include <algorithm>
#include <vector>

using namespace std;

#define SKETCH_OUT "sketch.out"
#define GT_OUT "groundt.out"

#define PERMAT	6

double CurrentTime()
{
  struct timeval tv;
  if( gettimeofday(&tv, 0) < 0 ) {
    cerr << "Could not compute time of day" << endl;
    exit(1);
  }

  return double(tv.tv_sec) + double(tv.tv_usec)/1e6;
}

void ReadConfig(const char *configFileName,
		char *in_file1Name, char *in_file2Name,
		list<unsigned int> &threshLst,
		list<unsigned int> &rLst,
		list<unsigned int> &hLst,
		list<unsigned int> &inxLst,
		list<int> &iterLst)
{
  cout << "Reading the configuration file..." << endl;

  ifstream configFile(configFileName, ios::in);

  if( !configFile ) {
    cerr << "Could not open the configuration file" << endl;
    exit(1);
  }

  char cmd;
  int temp = 0;

  while( configFile >> cmd ) {
    switch( cmd ) {
    case 'f':
      configFile >> in_file1Name >> in_file2Name;
      break;
    case 't':
      configFile >> temp;
      threshLst.push_back(temp);
      break;
    case 'r':
      configFile >> temp;
      rLst.push_back(temp);
      break;
    case 'h':
      configFile >> temp;
      hLst.push_back(temp);
      break;
    case 'i':
      configFile >> temp;
      inxLst.push_back(temp);
      break;
    case 's':
      configFile >> temp;
      iterLst.push_back(temp);
      break;
    default:
      cerr << "Invalid config command: " << cmd << " not recognized" << endl;
      exit(1);
    }
  }   
}

void CompareData(const list<unsigned int> gtKeyList,
		 const list<unsigned int> sketchKeyList,
		 ofstream &outFile)
{
  int falsePos = 0;
  int truePos = 0;
  int falseNeg = 0;

  int expected = gtKeyList.size();

  list<unsigned int>::const_iterator gtIt = gtKeyList.begin();
  list<unsigned int>::const_iterator sketchIt = sketchKeyList.begin();

  while( true ) {
    if( (gtIt == gtKeyList.end()) && (sketchIt == sketchKeyList.end()) )
      break;
    
    if( gtIt == gtKeyList.end() ) {
      falsePos++;
      sketchIt++;
    }
    else if( sketchIt == sketchKeyList.end() ) {
      falseNeg++;
      gtIt++;
    }
    else if( *sketchIt < *gtIt ) {
      falsePos++;
      sketchIt++;
    }
    else if( *gtIt < *sketchIt ) {
      falseNeg++;
      gtIt++;
    }
    else {
      truePos++;
      gtIt++;
      sketchIt++;
    }
  }
   
  outFile << "Expected_True_Positives: \t" << expected << endl;
  
  outFile << "True_Positives: \t" << truePos << "\t"
	  << ((double)truePos / (double)expected)*100 << "%" << endl;
  
  outFile << "False_Negatives: \t" << falseNeg << "\t"
	  << ((double)falseNeg / (double)expected)*100 << "%" << endl;

  outFile << "False_Positives: \t" << falsePos << "\t"
	  << ((double)falsePos / (double)expected)*100 << "%" << endl;

  outFile << endl
	  << "-------------------------------------------------------"
	  << endl;
}

void ProcessFiles(char *outFileName)
{
  ifstream groundtFile(GT_OUT, ios::in);

  if( !groundtFile ) {
    cerr << "Ground truth output could not be found or opened" << endl;
    exit(1);
  }

  list<unsigned int> gtKeyList;
  unsigned int key, change;

  while( groundtFile >> key >> change )
    gtKeyList.push_back(key);

  gtKeyList.sort();


  ifstream sketchFile(SKETCH_OUT, ios::in);
  
  if( !sketchFile ) {
    cerr << "Sketch output could not be found or opened" << endl;
    exit(1);
  }

  int thresh, r, numTbls, inxBits, iterSize;
  unsigned int Itemnum;
  double insertTime, cdTime;
  sketchFile >> thresh;
  sketchFile >> r;
  sketchFile >> numTbls;
  sketchFile >> inxBits;
  sketchFile >> iterSize;
  sketchFile >> insertTime;
  sketchFile >> cdTime;
  sketchFile >> Itemnum;

  list<unsigned int> sketchKeyList;
  while( sketchFile >> key )
    sketchKeyList.push_back(key);

  sketchKeyList.sort();


  ofstream outFile(outFileName, ios::app);

  if( !outFile ) {
    cerr << "Output file could not be opened" << endl;
    exit(1);
  }

  outFile << "Threshold: \t"      << thresh << endl
	  << "R: \t"              << r      << endl
	  << "Num_Tables: \t"     << numTbls << endl
	  << "Num_Inx_Bits: \t"   << inxBits << endl 
	  << "Iteration_Size: \t" << iterSize << endl
	  << "Insertion_Time: \t" << insertTime << "\ts" << endl
	  << "Detection_Time: \t" << cdTime << "\ts" << endl
	  << "Memory_Acess_times: \t" << (Itemnum*(numTbls*2+8)) <<endl
	  << "Updated_Stage_Fast_Memory_Usage_size: \t" << (8192+numTbls*16*(1U<<inxBits)) <<endl
          << endl;

  CompareData(gtKeyList, sketchKeyList, outFile);
}

void GetThresholds(list<unsigned int> &threshLst, 
		  char *in_file1, char *in_file2)
{
  char cmd[512];
  unsigned int max = 0;
  unsigned int thresh = 1000000;

  list<unsigned int>::iterator listIt;
  // Find the largest number of anomalies
  for( listIt = threshLst.begin(); listIt != threshLst.end(); listIt++ )
    if( *listIt > max ) max = *listIt;
  
  vector<unsigned int> gtChangeVec;

  cout << "Looking for thresholds..." << endl;

  do {
    cout << "\tTrying threshold value " << thresh << ", ";
    gtChangeVec.clear();

    sprintf(cmd, "./groundt %s %s %d > %s",
	    in_file1, in_file2, thresh, GT_OUT);

    system(cmd);
    
    ifstream groundtFile(GT_OUT, ios::in);

    if( !groundtFile ) {
      cerr << "Ground truth output could not be found or opened" << endl;
      exit(1);
    }
    
    unsigned int key, change;

    while( groundtFile >> key >> change )
      gtChangeVec.push_back(change);
    
    cout << "found " << gtChangeVec.size() << " anomalies" << endl;
 
    thresh /= 5;

  } while( gtChangeVec.size() < (max + 1) );

  sort(gtChangeVec.begin(), gtChangeVec.end());

  list<unsigned int> threshVals;
  for( listIt = threshLst.begin(); listIt != threshLst.end(); listIt++ ) {
    int size = gtChangeVec.size();
    unsigned int val1 = gtChangeVec[size - *listIt - 1];
    unsigned int val2 = gtChangeVec[size - *listIt];
    threshVals.push_back((val1 + val2) / 2);
  }

  threshVals.sort();

  threshLst.clear();
  threshLst.merge(threshVals);
}

int main(int argc, char** argv) {
  if( argc != 3 ) {
    cerr << "Usage: out_parser <config file> <output file>" << endl;
    exit(1);
  }

  char cmd[512];

  char in_file1[256], in_file2[256];
  list<unsigned int> threshLst, rLst, hLst, inxLst;
  list<int> iterLst;

  ReadConfig(argv[1], in_file1, in_file2, threshLst, rLst, 
	     hLst, inxLst, iterLst);

  // Get the actual thresholds
  GetThresholds(threshLst, in_file1, in_file2);

  // Run each of the trials
  for( list<unsigned int>::iterator threshIt = threshLst.begin(); 
       threshIt != threshLst.end();
       threshIt++ ) {

    // RUN GROUNDT
    cout << "\nFinding ground truth with threshold " << *threshIt << endl;
    sprintf(cmd, "./groundt %s %s %d > %s", 
	    in_file1, in_file2, *threshIt, GT_OUT);
    
    if( system(cmd) != 0 ) {
      cerr << "Error running groundt" << endl;
      exit(1);
    }

    for( list<int>::iterator iterIt = iterLst.begin();
	 iterIt != iterLst.end();
	 iterIt++ ) {

      for( list<unsigned int>::iterator rIt = rLst.begin();
	   rIt != rLst.end();
	   rIt++ ) {

	for( list<unsigned int>::iterator hIt = hLst.begin();
	     hIt != hLst.end();
	     hIt++ ) {

	  for( list<unsigned int>::iterator inxIt = inxLst.begin();
	       inxIt != inxLst.end();
	       inxIt++ ) {
	  
	    cout << "\tIterations = " << *iterIt 
		 << ", R = " << *rIt 
		 << ", H = " << *hIt 
		 << ", Inx bits used = " << *inxIt << endl;

	    sprintf(cmd, "./sketch_test %s %s %d %d %d %d %d > %s",
		    in_file1, in_file2, *threshIt, *rIt, *hIt, 
		    *inxIt, *iterIt, SKETCH_OUT);

	    if( system(cmd) != 0 ) {
	      cerr << "Error running sketch_test" << endl;
	      exit(1);
	    }
	    
	    ProcessFiles(argv[2]);
	  }
	}
      }
    }
  }

  return 0;
}
