#include <math.h>
#include <iostream>
#include <vector>
#include <map>
#include "globals.h"
#include "revHash.h"
#include "hash.h"

using namespace std;

RevHashTable::RevHashTable(int seed, int inxBits)
{
  uint32 i;
  
  if( (inxBits % NUM_DIVISIONS) != 0 ) {
    cerr << "Number of index bits must be divisible by "
	 << NUM_DIVISIONS << endl;
    exit(-1);
  }

  numBins = (1U << inxBits);
  binMask = numBins - 1;
  bitsPerDiv = inxBits / NUM_DIVISIONS;
  binsPerDiv = (1U << bitsPerDiv);
  divMask = binsPerDiv - 1;
  
  // Initialize value table to all 0's
  valueTable = new ValueType[numBins];
  for ( i = 0; i < numBins; i++ )
    valueTable[i] = 0;
    
  for( i = 0; i < 8; i++ ) {
    revLookupTable[i].resize(MAX_KEY + 1);
    PopulateLookupTable(i, (seed+1) * (i+1));
  }

  // Allocate space for chgLst
  chgLst = new ValueType[numBins];
}

RevHashTable::~RevHashTable()
{
  delete[] valueTable;
  delete[] chgLst;
}

void RevHashTable::PopulateLookupTable(int div, int seed)
{
  // Initialize the random number generator
  srand(seed + 1);

  for( uint32 i = 0; i <= MAX_KEY; i++ ) {
    uint32 bin = uint8(double(binsPerDiv)*rand()/(RAND_MAX+1.0));

    revLookupTable[div][bin].push_back(uint8(i));

    // shift bin so that in GetIndex() we can OR things together
    lookupTable[div][i] = (bin << (div*bitsPerDiv));
  }
}

void RevHashTable::GetPossibleKeys(list<uint8> &keys, int div, IndexType bin)
    {
      bin = (bin >> (div * bitsPerDiv)) & divMask;

      vector<uint8> &rt = revLookupTable[div][bin];
  
      keys.clear();
      int numKeys = rt.size();
      for( int i = 0; i < numKeys; i++ )
        keys.push_back(rt[i]);
    }

void RevHashTable::ChangeDetection(list<IndexType> &changes,
				   const RevHashTable *otherTbl, 
				   double threshold)
{
  changes.clear();
  
  ValueType difference;
  for( uint32 i = 0; i < numBins; i++ ) {
    difference = valueTable[i] - otherTbl->valueTable[i];
#ifdef ABSOLUTE_CHANGE
    if( difference < 0 ) difference = -difference;
#endif
    if( difference > threshold ) {
      //cerr << "\tbkt: " << i << " = \t" << difference << endl;
      changes.push_back(i);
    }
  }
}

static int compare_value(const void* v1, const void* v2)
{
  ValueType x1 = *(ValueType*)v1;
  ValueType x2 = *(ValueType*)v2;

  if ( x1 > x2 )
    return -1;
  else if ( x1 < x2 )
    return 1;
  else
    return 0;    
}

double RevHashTable::FindThreshold(const RevHashTable *otherTbl,
				   int numAnoms, double minThreshold)
{
  ValueType difference;

  if( numAnoms == -1 )
    return minThreshold;

  // Create a map with a key of the change mag. and a value of the corr. bin
  int numChgs = 0;

  for( uint32 i = 0; i < numBins; i++ ) {
    difference = valueTable[i] - otherTbl->valueTable[i];
#ifdef ABSOLUTE_CHANGE
    if( difference < 0 ) difference = -difference;
#endif
    if( difference > minThreshold ) {
      chgLst[numChgs++] = difference;
    }
  }

  if( numChgs == 0 )
    return -1.0;
  else if( numChgs <= numAnoms )
    return minThreshold;
  else {
    // Sort the list of changes in decreasing order
    qsort(chgLst, numChgs, sizeof(ValueType), compare_value);
    return chgLst[numAnoms-1] - 1;
  }
}


