#include <math.h>
#include <map>
#include <set>
#include "revKSketch.h"
#include "mangler.h"
#include "globals.h"

#define EXTEND_RATIO	0.9

using namespace std;

void ListIntersect(list<IndexType> &outList,
                   const list<IndexType> &l1,
                   const list<IndexType> &l2);

RevKSketch::RevKSketch(int numHashTbls, int myR, int inxBits, int mySize, int OKSTbls, int OKSbits)
{
    numTbls = numHashTbls;
    r = myR;
    iterSize = mySize;

    extraSketch = new KSketch(OKSTbls,OKSbits);

    // Populate data tables
    hashTbls = new RevHashTable*[numTbls];
    medBuf = new ValueType[numTbls];
    for( int i = 0; i < numTbls; i++ )
    {
        hashTbls[i] = new RevHashTable(i, inxBits);
        medBuf[i] = 0;
    }

    uint32 K = hashTbls[0]->GetNumBins();
    invFactor = 1.0 / (K - 1);

#ifdef NEW_MANGLE

    mangler = new Mangler();
#else

    a = A;
    a_inv = FindInverse(a);
#endif
}

RevKSketch::~RevKSketch()
{
#ifdef NEW_MANGLE
    delete mangler;
#endif

    delete extraSketch;
    for ( int i = 0; i < numTbls; i++ )
        delete hashTbls[i];
    delete[] hashTbls;
    delete[] medBuf;
}

void RevKSketch::ChangeDetection(list<KeyType> &resultList,
                                 RevKSketch &otherSketch, double threshold)
{
    bool done = false;

    // Get the heavy change buckets for each table
    vector<list<IndexType> > bktLists(numTbls);
    int iter = 0;

    while( !done && (iter < MAX_ITER) )
    {
        iter++;

        int tbl;
        ValueType nextThresh;
        ValueType maxThresh = EXTEND_RATIO*threshold;


        // Find a threshold (maxThresh) such that no table will return
        // more than ITER_ANOM anomalies
        for( tbl = 0; tbl < numTbls; tbl++ )
        {
            bktLists[tbl].clear();

            nextThresh = hashTbls[tbl]->FindThreshold(otherSketch.hashTbls[tbl],
                         iterSize, maxThresh);

            if( nextThresh == -1.0 )
                done = true;
            else if( nextThresh > maxThresh )
                maxThresh = nextThresh;
        }

        // Get the list of buckets corresponding to the threshold determined
        // above.
        for( tbl = 0; tbl < numTbls && !done; tbl++ )
        {
            hashTbls[tbl]->ChangeDetection(bktLists[tbl],
                                           otherSketch.hashTbls[tbl],
                                           maxThresh);
        }

        // Perform an iteration of the key solution algorithm
        if( !done )
        {
            // Create the nodes in the graph
            GraphType graph;
            graph.clear();
            CreateNodes(graph, bktLists);

            // Create the links between each level
            for( int div = 0; div < 3; div++ )
                CreateLinks(graph, div);

            // Get the full keys from the graph after the traversal
            list<KeyType> tempLst;
            GenerateFullKeys(graph, tempLst);

            done = true;
            list<KeyType>::iterator keyIt;

            // Test each reported keys in the extra sketch
            for( keyIt = tempLst.begin(); keyIt != tempLst.end(); keyIt++ )
            {
                ValueType extraEst = extraSketch->EstValue(*keyIt);
                ValueType othExtraEst = otherSketch.extraSketch->EstValue(*keyIt);
#ifdef ABSOLUTE_CHANGE

                ValueType diff =
                    extraEst > othExtraEst ?
                    extraEst - othExtraEst :
                    othExtraEst - extraEst;
#else

                ValueType diff = extraEst - othExtraEst;
#endif

		// for the verifier sketch, check if the difference is significant
                //if( diff > maxThresh)
		// original code, compared with the final threshold instead of the iterative threshold
                 if( diff > threshold )
                {
                    resultList.push_back(*keyIt);
                    RemoveValue(*keyIt, extraEst);
                    otherSketch.RemoveValue(*keyIt, othExtraEst);
                    done = false;  // Found another key, so keep going
                }
            }
        }
    }

    resultList.sort();
    resultList.unique();
}

void RevKSketch::CreateNodes(GraphType &graph,
                             const vector<list<IndexType> > &bktLists)
{
    // A map that associates key segments with nodes
    map<unsigned char, NodeType*> nodeMap;

    graph.resize(4);  // Initialize the graph

    // For each key division
    for( int div = 0; div < 4; div++ )
    {
        nodeMap.clear();

        // For each hash table in the sketch
        for( int tbl = 0; tbl < numTbls; tbl++ )
        {

            // For each heavy change bucket for this hash table
            for( list<IndexType>::const_iterator bktListIt = bktLists[tbl].begin();
                    bktListIt != bktLists[tbl].end();
                    bktListIt++ )
            {

                list<unsigned char> keys;
                hashTbls[tbl]->GetPossibleKeys(keys, div, *bktListIt);

                // For each of the possible key segments
                for( list<unsigned char>::iterator keyIt = keys.begin();
                        keyIt != keys.end();
                        keyIt++ )
                {

                    NodeType *node = nodeMap[*keyIt];  // Get the node, if one exists

                    if( node == NULL )
                    { // Key does not already exist in the map
                        // Create a new node and add it to the map
                        node = new NodeType(*keyIt, *bktListIt, tbl, numTbls);
                        nodeMap[*keyIt] = node;
                    }
                    else
                    {
                        if( node->bktLists[tbl].empty() ) // First entry for this HT
                            node->numFound++;

                        node->bktLists[tbl].push_back(*bktListIt);
                    }
                }  // End for ... keyIt
            }  // End for ... bktListIt
        }  // End for ... tbl

        // Enter the nodes where numTbls - numFound <= R into the graph
        for( map<unsigned char, NodeType*>::iterator nodeMapIt = nodeMap.begin();
                nodeMapIt != nodeMap.end();
                nodeMapIt++ )
        {
            NodeType *node = nodeMapIt->second;

            if( (numTbls - node->numFound) <= r )
                graph[div].push_back(node);
            else
                delete node;
        }  // End for ... nodeMapIt
    }  // End for ... div
}

void RevKSketch::CreateLinks(GraphType &graph, int startDiv)
{
    if( startDiv >= 3 )
        return;  // cannot start at the last level

    list<NodeType*> *l1List = &(graph[startDiv]);
    list<NodeType*> *l2List = &(graph[startDiv+1]);

    list<NodeType*> linkList;

    for( list<NodeType*>::iterator l1It = l1List->
                                          begin();
            l1It != l1List->end();
            l1It++ )
    {

        for( list<NodeType*>::iterator l2It = l2List->
                                              begin();
                l2It != l2List->end();
                l2It++ )
        {

            NodeType *newNode = RIntersect(*l1It, *l2It);
            if( newNode != NULL )
                linkList.push_back(newNode);
        }  // End for...l2It

        delete *l1It;  // no longer need this node in level 1
    }

    // Delete all of the nodes in L2
    for( list<NodeType*>
            ::iterator l2It = l2List->begin();
            l2It != l2List->end();
            l2It++ )
        delete *l2It;

    l1List->clear();
    l2List->clear();
    l2List->insert(l2List->end(), linkList.begin(), linkList.end());
}

NodeType* RevKSketch::RIntersect(NodeType *node1, NodeType *node2)
{
    NodeType *newNode = new NodeType;
    newNode->bktLists.resize(numTbls);

    for( int tbl = 0; tbl < numTbls; tbl++ )
    {
        list<IndexType> *node1Lst = &(node1->bktLists[tbl]);
        list<IndexType> *node2Lst = &(node2->bktLists[tbl]);
        list<IndexType> *newLst = &(newNode->bktLists[tbl]);

        ListIntersect(*newLst, *node1Lst, *node2Lst);

        if( newLst->empty() )
        {
            if( newNode->numFound >= r )
            {
                delete newNode;
                return NULL;
            }
            else
                newNode->numFound++;
        }
    }

    list<unsigned char> *keyLst = &(newNode->keyList);
    keyLst->insert(keyLst->end(), node1->keyList.begin(), node1->keyList.end());
    keyLst->insert(keyLst->end(), node2->keyList.begin(), node2->keyList.end());

    return newNode;
}

void RevKSketch::GenerateFullKeys(GraphType &graph, list<KeyType> &resultList)
{
    unsigned int fullKey;
    unsigned char *keyDivided = (unsigned char *)&fullKey;

    for( list<NodeType*>::iterator lstIt = graph[3].begin();
            lstIt != graph[3].end();
            lstIt++ )
    {

        list<unsigned char> *keyList = &((*lstIt)->keyList);
        if( keyList->size() == 4 )
        {
            int div = 0;
            for( list<unsigned char>::iterator keyIt = keyList->
                    begin();
                    keyIt != keyList->end();
                    keyIt++, div++ )
            {

                keyDivided[div] = *keyIt;
            }

            resultList.push_back(RestoreKey(fullKey));
        }  // end if

        delete *lstIt;
    }  // end for
}

void ListIntersect(list<IndexType> &outList,
                   const list<IndexType> &l1,
                   const list<IndexType> &l2)
{
    outList.clear();

    list<IndexType>::const_iterator l1It = l1.begin();
    list<IndexType>::const_iterator l2It = l2.begin();

    while( true )
    {
        if( (l1It == l1.end()) || (l2It == l2.end()) )
            return;

        if( *l1It > *l2It )
            l2It++;
        else if( *l1It < *l2It )
            l1It++;
        else
        {
            outList.push_back(*l1It);
            l1It++;
            l2It++;
        }
    }
}

#ifndef NEW_MANGLE
typedef struct ExtEuclidType
{
    long long d;
    long long x;
    long long y;

    ExtEuclidType(long long myD, long long myX, long long myY)
            : d(myD), x(myX), y(myY)
    {}

    ExtEuclidType()
    {}
}
ExtEuclidType;

ExtEuclidType ExtendedEuclid(long long a, long long b)
{
    if( b == 0 )
        return ExtEuclidType(a, 1, 0);

    ExtEuclidType extEuclidPrime = ExtendedEuclid(b, a % b);

    ExtEuclidType extEuclid;
    extEuclid.d = extEuclidPrime.d;
    extEuclid.x = extEuclidPrime.y;
    extEuclid.y = extEuclidPrime.x - (a / b)*extEuclidPrime.y;

    return extEuclid;
}

unsigned int RevKSketch::FindInverse(unsigned int a)
{
    return (unsigned int)((ExtendedEuclid(a, N)).x % N);
}
#endif

IndexType RevKSketch::GetIndex(KeyType key, int tbl)
{
    return hashTbls[tbl]->GetIndex(ReorderKey(key));
}
